//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_grid.C,v 1.59 2007/12/07 20:33:01 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_grid.C
//
#include "prom_grid.hh"
/*!  
  PromGridBase construictor
  \sa ~PromGridBase()
  \param opts Options data base
  \param perf Performance monitor object
  \param bank Bank for Parents (hyper edges)
  \param nloc number of nodes local to this processor
  \param nglob total number of nodes
  \param fact factor by which processor are reduced (ie, if ==1 then all processors are active
  \param wcomm global communicator for this level (eg, PETSC_COMM_WORLD)
  \param bsize Dof per node
*/ 
#undef __FUNCT__    
#define __FUNCT__ "PromGridBase::PromGridBase" 
PromGridBase::PromGridBase( const PromOptions &opts,
			    PromPerfMonitor &perf,
			    PromParentBank * const bank, 
			    const int nloc, 
			    const int fact, const PromComm &wcomm ) :
  factor_(fact), 
  nLocalNd_(nloc),
  nNodes_(0), 
  nBlocks_(0),
  blockISs_(NULL),
  gid_extraParents_(NULL),
  stiffness_(NULL),
  prolongator_(NULL),
  scalarMap_(NULL),
  CRmap_(NULL),
  options_(opts),
  perf_mon_(perf),
  parBank_(bank),
  ghosts_(nloc/3+617),
  CRAggs_(NULL),
  nCRAggs_(0),
  group_(MPI_GROUP_NULL),
  mpi_group_world_(MPI_GROUP_NULL),
  gComm_(MPI_COMM_NULL)
{
  BuildComunicator(wcomm);
}

#undef __FUNCT__    
#define __FUNCT__ "PromGridBase::BuildComunicator" 
int PromGridBase::BuildComunicator(const PromComm &wcom)
{
  // build communicator: 'gComm_'
  MPI_Comm cm, wcomm = wcom.MPIComm(); 
  if( factor_ == 1 ) {
    mpi_group_world_ = MPI_GROUP_NULL;
    group_ = MPI_GROUP_NULL;
    gComm_ = wcomm;
  }
  else { // factor != 1
    int nCommprocs_ = wcom.np()/factor_, ierr, ftag;
    MPI_Comm_group( wcomm, &mpi_group_world_ );
    int ranges[1][3] = { {0, factor_ * (nCommprocs_-1), factor_ } };
    MPI_Group_range_incl( mpi_group_world_, 1, ranges, &group_ );
    MPI_Comm_create( wcomm, group_, &cm );
    if( cm != MPI_COMM_NULL ) {
      ierr = PetscCommDuplicate( cm, &gComm_, &ftag );CHKERRQ(ierr);
      ierr = MPI_Comm_free( &cm );CHKERRQ(ierr);
      MPI_Errhandler_set( gComm_, MPI_ERRORS_RETURN );
    }
    else {
      gComm_ = MPI_COMM_NULL;
    }
  }
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGridBase::~PromGridBase" 
PromGridBase::~PromGridBase()
{ 
  assert(this!=NULL);
  
  if( group_ != MPI_GROUP_NULL ) {
    MPI_Group_free( &group_ ); group_ = MPI_GROUP_NULL;
    MPI_Group_free( &mpi_group_world_ ); mpi_group_world_ = MPI_GROUP_NULL; 
    if(gComm_ != MPI_COMM_NULL) MPI_Comm_free( &gComm_ ); // PetscCommDestroy(&gComm_); // -- causese a segv on linux???
  }
  ghosts_.removeAll();

  // parents for (dynamic) regularization nodes
  if( gid_extraParents_ != NULL ) {
    PromParentArr_extra *parr;
    while( !gid_extraParents_->isEmpty() ){
      parr = gid_extraParents_->removeHead();
      delete parr; 
    }
    delete gid_extraParents_; 
  }
  if( CRmap_ != NULL ) delete CRmap_;
  if( prolongator_ != NULL ) delete prolongator_;
  if( scalarMap_ != NULL ) delete scalarMap_ ;
  if( stiffness_ != NULL ) delete stiffness_;
  if( blockISs_ != NULL ) {
    assert( nBlocks_ > 0 ); 
    while( nBlocks_-- ){ 
      assert( blockISs_[nBlocks_] != NULL );
      PromISDestroy( blockISs_[nBlocks_] );  
    }
    PetscFree(blockISs_); //assert(ierr==0);
  }
  if( CRAggs_ != NULL ) PetscFree(CRAggs_);
}

/*! fine grid and normal grid constructor (const dof per node) */
#undef __FUNCT__    
#define __FUNCT__ "PromGrid::PromGrid" 
PromGrid::PromGrid(const PromOptions &opts, PromPerfMonitor &perf,
		   PromParentBank *const bank, const int nloc, 
		   const PromComm &wcomm, const int bsize ) :
  PromMap( nloc, bsize, wcomm ),
  PromBank(), 
  PromGridBase(opts,perf,bank,nloc,1,wcomm),
  nodes_(NULL),
  prevgrid_( NULL ),  
  nextgrid_( NULL )
{
}

/*! reduced coarse grid (const dof per node) */
PromGrid::PromGrid(const PromOptions &opts, PromPerfMonitor &perf,
		   PromParentBank*const bank, const int nloc, 
		   const int fact, const PromComm &wcomm, 
		   const int bsize ) :
  PromMap( (wcomm.mype()%fact==0) ? nloc : 0, bsize, wcomm ),
  PromBank(), 
  PromGridBase(opts,perf,bank,nloc,fact,wcomm),
  nodes_(NULL), prevgrid_( NULL ), nextgrid_( NULL )
{
}

/*! fine grid and normal grid constructor */
#undef __FUNCT__    
#define __FUNCT__ "PromGrid::PromGrid"
PromGrid::PromGrid( const PromOptions &opts, PromPerfMonitor &perf,
		    PromParentBank*const bank, const int nloc, 
		    const PromComm &wcomm, const int lid_ndof[] ) :
  PromMap( nloc, lid_ndof, wcomm ),
  PromBank(), 
  PromGridBase(opts,perf,bank,nloc,1,wcomm),
  nodes_(NULL),
  prevgrid_( NULL ),  
  nextgrid_( NULL )
{
}

/*! reduced coarse grid */
PromGrid::PromGrid( const PromOptions &opts, PromPerfMonitor &perf, 
		    PromParentBank * const bank, const int nloc,
		    const int fact, const int nLocPartEq, const PromComm &wcomm ) :
  PromMap((wcomm.mype()%fact==0) ? nloc : 0, wcomm, // hack for VBR  
	  (wcomm.mype()%fact==0) ? nLocPartEq : 0 ),
  PromBank(), 
  PromGridBase(opts,perf,bank,nloc,fact,wcomm),
  nodes_(NULL),  prevgrid_( NULL ), nextgrid_( NULL )
{
}

/*!  helpers */
bool PromGrid::isActive()const
{ 
  assert(this!=NULL); assert(factor_>0);
  return (comm_.mype()%factor_ == 0);
}
bool PromGrid::isActive( int proc )const
{
  assert(this!=NULL); return (proc%factor_==0);
}
int PromGrid::getLevel()const
{  
  int ii = 0;
  for( PromGrid *g = prevgrid_ ; g != NULL ; g = g->prevgrid_ ) ii++;
  return ii;
}
int PromGrid::getNumLevels()const
{
  int ii = getLevel(); 
  for( PromGrid *g = nextgrid_; g != NULL ; g = g->nextgrid_ ) ii++;
  return ii + 1;
}

int  PromGrid::FineGridInitWithA() 
{
  return 0; 
}

#undef __FUNCT__
#define __FUNCT__ "PromGrid::InitNodes" 
/* PromGrid::InitNodes ******************************************
 *
 */  
int PromGrid::InitNodes( const int nnodes )
{
  assert(this!=NULL);
  assert(nodes_ == NULL);  assert(nNodes_ == 0); assert(nnodes >= nLocalNd_);
  if( nnodes > 0 ) {
    // simple initialization
    int ierr = AddNNodePointers( nnodes ); CHKERRQ(ierr); 
    // set nodes data strictures
    for( int i = 0 ; i < nnodes; i++ ) {
      PromNode *curr = newNode( i );
      nodes_[i] = curr;
    }
  }
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGrid::AddNNodePointers" 
/* PromGrid::AddNNodePointers ******************************************
 *
 */  
int PromGrid::AddNNodePointers( int nn )
{
  if( nn > 0 ) {
    PromNode **newNodes; int ierr, ii, newnnodes = nn+nNodes_; 
    ierr = PetscMalloc(sizeof(PromNode*)*(nn+nNodes_+1),&newNodes);
    CHKERRQ(ierr); 
    for( ii = 0 ; ii < nNodes_ ; ii++ ) newNodes[ii] = nodes_[ii]; //cp old
    do newNodes[ii] = NULL; while( ii++ < newnnodes );            // clear new
    if( nodes_ != NULL ) PetscFree( nodes_ );
    nodes_ = newNodes;
    nNodes_ = newnnodes;
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::~PromGrid" 
PromGrid::~PromGrid()
{
  if( nodes_ ) {
    PetscFree( nodes_ ); nodes_ = NULL;
  }
}

int PromGrid::masterProc() const
{
  assert(this!=NULL); 
  return (comm_.mype()/factor_)*factor_;
}

int PromGrid::masterProc( const int p ) const
{ 
  assert(this!=NULL);
  return (p/factor_) * factor_; 
}

int PromGrid::getEndProc()const
{
  assert(this!=NULL);
  int ret =  masterProc() + factor_;
  return (ret>comm_.np()) ? comm_.np() : ret;
}

int PromGrid::getEndProc( const int p ) const
{
  assert(this!=NULL);
  int ret =  masterProc(p) + factor_;
  return (ret>comm_.np()) ? comm_.np() : ret;
}

bool PromGrid::isSameGroup( int proc1, int proc2 ) const
{ 
  assert(this!=NULL);
  return ( (proc1/factor_)*factor_ == (proc2/factor_)*factor_ );
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::MakeMatrix"
/* PromGrid::MakeMatrix ************************************
 *
 *  - Create stiffness matrix
 *
 *  INPUT:
 *   - CR_id: CR matrix for regularization
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code 
 */
int PromGrid::MakeMatrix( bool useComplex/*=false*/,  bool useCRmat/*=false*/, 
			  const int extras/*=0*/, 
			  PromTable<PromCR*> *CR_id /*= NULL*/,
			  void *mat /*= NULL*/ )
{   
  assert(this!=NULL); TablePos tpos;
  int ierr,xx,*row_nnz=NULL,*offnnz=NULL,nloc=nlocal(),noffmax=nglobal()-nloc;

  PetscFunctionBegin;  
  if( nloc > 0 && mat == NULL ) {
    int nl,ng,self = nloc + 1;
    ierr = PetscMalloc(self*sizeof(int), &row_nnz );   CHKERRQ(ierr);
    PetscMemzero( row_nnz, self * sizeof(int) );      
    ierr = PetscMalloc(self*sizeof(int), &offnnz);     CHKERRQ(ierr);
    PetscMemzero( offnnz, self * sizeof(int) );
    row_nnz[nloc] = offnnz[nloc] = -9999; 
    if( nodes_[0]->symbData_==NULL)SETERRQ(1,"symbData==NULL");
    for( xx = 0 ; xx < nloc ; xx++ ) {
      PromNode *curr = nodes_[xx]; assert(curr->symbData_ != NULL);
      // count degree for latter matrix creation
      PromTable<PromNode*> *tab = &curr->symbData_->adjac;
      tpos = tab->getHeadPosition(); nl = 1; ng = self = 0; 
      while( tpos ) { 
	PromNode* adj = tab->getNext( tpos ); 
	if( curr == adj ) self++;
	else if( adj->isGhost() ) ng++;  
	else nl++;
      }
      assert(self == 0);
      assert(nl > 0); // include self
      row_nnz[xx] = nl + extras; 
      if( row_nnz[xx] > nloc ) row_nnz[xx] = nloc;
      offnnz[xx] = ng + extras; 
      if( offnnz[xx] > noffmax ) offnnz[xx] = noffmax;
    }
  }

  // add regularization edges -- create
  if( stiffness_ != NULL ) SETERRQ(1,"stiffness is here!!!");
  if( CR_id != NULL && mat == NULL ) {
    assert(!useCRmat);
    int my0 = my0Nd(), next0 = my0 + nlocal(), ii, jj, proc;
    const int *p_gnode = proc_gnode(), np = comm_.np(), mype = comm_.mype();
    const int reqsz = sizeof(MPI_Request)/sizeof(int); 
    MPI_Comm wcomm = comm_.MPIComm(); const int tag = comm_.getNewTag();
    PromList<MPI_Request*> smessages; 
    MPI_Status status; 
    PromTable<PromList<int>*> proc_list(57); PromList<int> *list; 
    int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 

    // collect sends
    TablePos tpos = CR_id->getHeadPosition();
    while( tpos ) {
      PromCR* mcr = CR_id->getNext( tpos );
      int *pi = mcr->gnodes_, len = mcr->len();
      // get local nodes 
      for( ii = 0 ; ii < len ; ii++ ) {
	int gidi = pi[ii];
	if( gidi < my0 || gidi >= next0 ) { // off proc -- send
	  PromGrid::FindProc(p_gnode, np, gidi, proc); assert(proc!=mype);
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromList<int>(100); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->addTail( gidi+1 ); list->addTail( len + 1 ); 
	  for(jj=0;jj<len;jj++)list->addTail( pi[jj] + 1 ); 
	}
	else if( row_nnz != NULL ) { // local -- add
	  int lid = gidi - my0; 
	  for(jj=0;jj<len;jj++){ 
	    int gidj = pi[jj];
	    if( gidj < my0 || gidj >= next0 ) offnnz[lid]++;
	    else row_nnz[lid]++;
	  }
	}
      }
    }
    // send(tag) *< gid, nj, nj*<gidj> >
    for( ii = 0 ; ii < np ; ii++ ) sbuf[ii] = 0;  
    while( !proc_list.isEmpty() ) {
      list = proc_list.removeHead( &proc ); proc--;
      int tt = list->getCount(); 
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
      CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff;
      while( !list->isEmpty() ) *pb++ = list->removeHead() - 1;  
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      // send(tag) *< gid, nj, nj*<gidj> >
      MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
      delete list;
      sbuf[proc] = tt; assert(tt>0);
    }

    // recive requests -- global!
    MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, wcomm );
    for( proc = 0 ; proc < np ; proc++ ) {
      if( rbuf[proc] != 0 ) {
	// recv(tag)
	int tt = rbuf[proc], *buff, *pb; 
	ierr = PetscMalloc(tt*sizeof(int),&buff );  CHKERRQ(ierr);
	pb = buff;
	// recv(tag) *< gid, nj, nj*<gidj> >
	MPI_Recv( buff, tt, MPI_INT, proc, tag, wcomm, &status );
	MPI_Get_count( &status, MPI_INT, &ierr ); assert(tt==ierr);
	proc = status.MPI_SOURCE;
	while( pb-buff < tt ) {
	  int gidi = *pb++; assert(gidi >= my0 && gidi < next0);
	  int len = *pb++;
	  for( ii = 0 ; ii < len ; ii++ ) {
	    int gidi = pb[ii];
	    if( gidi >= my0 && gidi < next0 && row_nnz != NULL ) {
	      int lid = gidi - my0; 
	      for(jj=0;jj<len;jj++){ 
		int gidj = pb[jj];
		if( gidj < my0 || gidj >= next0 ) offnnz[lid]++;
		else row_nnz[lid]++;
	      }
	    }
	  }
	  pb += len;
	}
	assert(pb-buff == tt); 	PetscFree(buff);
      }
    }
    // cleanup sends
    while( !smessages.isEmpty() ) {
      MPI_Request *sreq = smessages.removeHead();
      MPI_Wait( sreq, &status );    PetscFree( sreq );
    }
    MPI_Iprobe(MPI_ANY_SOURCE,tag,gComm_,&ii,&status);  assert(!ii);

    // check for too much
    for( xx = 0 ; xx < nloc && row_nnz != NULL ; xx++ ) {
      if( row_nnz[xx] > nloc ) row_nnz[xx] = nloc;
      if( offnnz[xx] > noffmax ) offnnz[xx] = noffmax;
    }
  }
  
  // make matrix
  Mat Amat = (Mat)mat;
  if( useCRmat ) {
    if( Amat != NULL ) stiffness_ = new PromCRMatrix( *this, Amat );
    else stiffness_ = new PromCRMatrix( this, 0, row_nnz, 0, offnnz );
    if( !stiffness_->isOK() ) SETERRQ(1,"new CR matrix failed"); 
  }
  else if( useComplex ) {
    assert(mat == NULL);
    stiffness_ = new PromComplexMatrix( this, 0, row_nnz, 0, offnnz );
    if( !stiffness_->isOK() ) SETERRQ(1,"new complex matrix failed"); 
  }
  else {
    if( Amat != NULL ) stiffness_ = new PromMatrix( *this, Amat );
    else stiffness_ = new PromMatrix( this, 0, row_nnz, 0, offnnz );
    if(!stiffness_->isOK()) SETERRQ(1,"new (regular) matrix failed");
  }
  //
  if( row_nnz != NULL ) PetscFree( row_nnz );
  if( offnnz != NULL ) PetscFree( offnnz );
  ierr = stiffness_->SetOption( MAT_KEEP_NONZERO_PATTERN);        CHKERRQ(ierr); 
  ierr = stiffness_->SetOption( MAT_STRUCTURALLY_SYMMETRIC );  CHKERRQ(ierr); 

  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::GetProc" 
/* PromGrid::GetProc ********************************************************
 *
 *  - find proc for global index
 *
 *  INPUT: 
 *    -  gid:
 *    -  index of this grid (out)
 *
 *  RETURN:
 *   - PETSc error code
 */
int PromGrid::GetProc( const int gid, int &proc ) const
{
  //bool ok = (gid >=0 && gid < proc_gnode_[comm_.np()]);
  //if( !ok ){ SETERRQ(1,"gid >=0 && gid < proc_gnode[numprocs]"); }
  const int np = comm_.np(), nglob = nglobal();
  float fproc = (float)gid * (float)np / (float) nglob + 0.5;
  proc = (int)fproc; 

  if( proc > np ) proc = np - 1;
  while(gid < proc_gnode_[proc]) proc--; 
  while(gid >= proc_gnode_[proc+1]) proc++;
  if(proc<0 || proc>=np){SETERRQ(1,"proc < 0 || proc >= numprocs");}
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::FindProc"
/* PromGrid::FindProc *************************************************
 *
 */
int PromGrid::FindProc(const int *p_g, const int np, const int gid, int &proc)
{
  int nnodes = p_g[np]; unsigned long gid_np = gid*np;
  //assert(gid >= 0 && gid < nnodes && np > 0 && p_g[0] == 0);
  proc = (int)( (float)(gid_np)/(float)nnodes + .5);

  if( proc > np ) proc = np - 1;
  while(gid < p_g[proc]) proc--; 
  while(gid >= p_g[proc+1]) proc++;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::Archive" 
/* PromGrid::Archive ***************************************************
 *
 *  - write grids to files (one for each proc)
 *
 *  INPUT: 
 *    -  file: file to read/write
 *    -  type: READ, WRITE, PRINT(pretty debug output)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */   
int PromGrid::Archive( FILE *file, const PromArchiveType type )
{
  int ierr, ii;
  
  if( type == PROM_PRINT ) {
    fprintf( file, "PromGrid:nN=%d,nLoc=%d,nGlob=%d",
	     nNodes_,nLocalNd_,nglobal() );
  }
  else if( type == PROM_WRITE ) {
    fprintf(file, "PromGrid: nN=%d,nLoc=%d,nGlob=%d,fact=%d\n", 
	    nNodes_, nLocalNd_, nglobal(), factor_);
    ierr=PromContext::printArr(file,proc_gnode_,comm_.np()+1,"proc_gnode[%d]");
    CHKERRQ(ierr);
  }
  else{
    int nl,nn,ng,fact,kk;
    kk = fscanf(file,"PromGrid: nN=%d,nLoc=%d,nGlob=%d,fact=%d\n", 
		&nn, &nl, &ng, &fact);
    if(nn != nNodes_ || nl != nLocalNd_ || ng != ng || fact != factor_||kk!=5){
      SETERRQ(1,"Wrong data");
    }
    assert( type == PROM_READ );
    ierr = PromContext::scanArr( file, &proc_gnode_, &ii, "proc_gnode[%d]");
    CHKERRQ(ierr);
    if ( ii != comm_.np()+1 ) SETERRQ(1,"Wrong number of procs"); 
  }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::NumSelected" 
/* PromGrid::NumSelected*****************************************************
 *
 *  - count selected nodes - called before copied up
 *
 *  INPUT:
 *    - nxt_nloc: (out)
 *    - nxt_nnodes: (out)
 *    - nxt_neq: num of local eqs
 *  
 *  RETURN:
 *   - PETSc error code 
 */
int PromGrid::NumSelected( int &nxt_nloc, int &nxt_nnodes, int &nxt_neq ) const
{
  int xx, ndf;

  for( nxt_nloc = nxt_nnodes = nxt_neq = xx = 0 ; xx < nNodes_; xx++ ) {
    PromNode* fine = nodes_[xx]; 
    if( fine->getState() == PROM_SELECTED ) {
      nxt_nnodes++;
      if( !fine->isGhost() ){ 
	nxt_nloc++; 
	ndf = fine->getNextNDF(); 
	nxt_neq += ndf;
      }
    }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::CreateProlongator"
/* PromGrid::CreateProlongator ***********************************************
 *   
 *  - construct prolongator (from 'prevG')
 *  - collective
 *
 *  INPUT:
 *    - prevG:
 *    - promcomm:
 *
 *  SIDE EFFECTS:
 * 
 *  RETURN:
 *    - PETSc error code
 */    
int PromGrid::CreateProlongator( PromGrid * const prevG, 
				 const PromComm &promcomm )
{
  int ierr, *nnz = NULL, *onnz = NULL; 
  int colLocalSZ = nlocalEq();
  int colGlobalSZ = nglobalEq();
  int rowLocalSZ = prevG->nlocalEq();

  // PROLONGATOR MATRIX: create & assembly of R    
  PetscFunctionBegin;
  ierr = PetscMalloc( (rowLocalSZ+1)*sizeof(int), &nnz);  CHKERRQ(ierr);
  ierr = PetscMalloc( (rowLocalSZ+1)*sizeof(int), &onnz); CHKERRQ(ierr);

  // create graph
  ierr = MakeProlongator( prevG, NULL, true, nnz, onnz ); CHKERRQ(ierr);

  // make scalar matrix for prolongator
  assert(prolongator_ == NULL);
  if( prevG->scalarMap_ == NULL ) {
    prevG->scalarMap_ = new PromMap( rowLocalSZ, 1, promcomm );
  }

  // factory (yuck)
  const PromMap &smap = *prevG->scalarMap_; 
  PromCRMatrix *crp = dynamic_cast<PromCRMatrix*>(prevG->stiffness_);
  if( crp != NULL ) {
    prolongator_ = new PromCRMatrix(smap,0,nnz,0,onnz,colLocalSZ,colGlobalSZ);
  }
  else if( !prevG->stiffness_->isComplex() ) {
    prolongator_ = new PromMatrix(smap,0,nnz,0,onnz,colLocalSZ,colGlobalSZ);
  }
  else {
    prolongator_ = new PromComplexMatrix(smap,0,nnz,0,onnz,colLocalSZ,colGlobalSZ);
  }
  if( prolongator_==NULL || !prolongator_->isOK() ) {
    SETERRQ3(1,"new PromMatrix failed rowLocalSZ=%d prolongator=%p",
	     rowLocalSZ,colLocalSZ,prolongator_);
  }
  if( nnz != NULL ) PetscFree( nnz ); 
  if( onnz != NULL ) PetscFree( onnz ); 

  // 
  //ierr = prolongator_->SetOption( MAT_ROWS_SORTED ); CHKERRQ(ierr);
  //ierr = prolongator_->SetOption( MAT_COLUMNS_SORTED ); CHKERRQ(ierr);
  ierr = prolongator_->SetOption( MAT_IGNORE_OFF_PROC_ENTRIES ); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__   
#define __FUNCT__ "PromGrid::MakeProlongator" 
/* PromGrid::MakeProlongator *****************************************
 *
 *  - create and fill ( R_{last}^{this} )
 *  - collective when nnz==NULL for assembly
 * 
 *  INPUT:
 *    - prevG:
 *    - prol: can be NULL (out)
 *    - clearDAP: 
 *    - nnz: can be NULL (out)
 *    - onnz: can be NULL (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */  
int PromGrid::MakeProlongator(const PromGrid *const prevG, PromMatrix *prol, 
			      const bool clearDAP /*=true*/, int *nnz/*=0*/,
			      int *onnz/*=0*/)
{
  int        ierr,k,kk,ii,jj,xx,leq;
  const int  nnleq = prevG->nlocalEq();
  const int  myn0 = my0Nd(), mypn10 = myn0 + nlocal();
  bool       getnnz = (nnz!=NULL), aggmg = (getMGType()==PROM_AGGMG);
  assert(!getnnz || prol==NULL);
  
  PetscFunctionBegin;
  bool flag = false; double tt = 0.0;
  ierr = options_.GetOptionDouble("-prometheus_filter",&tt,&flag);
  CHKERRQ(ierr);
  const double tol = flag ? tt : 1.e-12;

  // set values in matrix
  if( prevG->isActive() && (prol != NULL || getnnz) ) {
    if( getnnz ){
      // clear buffer
      for( ii=0 ; ii < nnleq ; ii++) nnz[ii] = onnz[ii] = 0;
    }
    // for all nodes on last level (cols)
    for( xx = leq = 0 ; xx < prevG->nLocalNd_  ; xx++ ){
      PromNode *finei = prevG->nodes_[xx];
      const int fndf = finei->getNDF();  
      PromParent * const *parI = finei->parents_->data_; 
      assert( parI != NULL );// global indices
      for( int JJ = 0 ; JJ < finei->nParents() ; JJ++ ){
	// R(ii,gidJ)
	const PromParent * const parJ = parI[JJ];
	int colj = parJ->geq(), rowi = finei->geq();
	const int gidJ = parJ->global(), cndf = parJ->getNDF(); assert(cndf>0);
	if( aggmg ) {
	  // add to matrix
	  shpfloat tmp, *shp;
	  if( parJ->DAP_ != NULL ) shp = parJ->DAP_;
	  else {
	    if(parJ->getPsz()==1){ tmp = parJ->getScalarP0(); shp = &tmp;}
	    else shp = parJ->P0_;
	  }
	  if( shp != NULL ) {
	    for( ii = k = 0 ; ii < fndf ; ii++, rowi++ ){
	      if( finei->isFixed(ii) ) k += cndf;
	      else if( getnnz ) {
		if( gidJ < myn0 || gidJ >= mypn10 ) onnz[leq+ii] += cndf;
		else nnz[leq+ii] += cndf; 
	      }
	      else {
		int cols[PROM_MAXNDF]; double vals[PROM_MAXNDF],t;
		for( jj = kk = 0 ; jj < cndf ; jj++, k++ ){
		  if( fabs( (t=shp[k]) ) > tol ) { // filter
		    cols[kk] = colj + jj; vals[kk] = t; kk++;
		  }
		} 
		if( kk > 0 ) {
		  ierr = prol->SetValues(1,&rowi,kk,cols,vals,INSERT_VALUES);
		  CHKERRQ(ierr);
		}
	      }
	    }
	  }
	}
      	else { // geometric R
	  double val = !getnnz ? parJ->getScalarP0() : 0.0;
	  for( k = 0 ; k < fndf ; k++, rowi++, colj++ ){
	    if( finei->isFixed(k) ) continue; // fine grid done manually 
	    if( getnnz ){
	      if( gidJ < myn0 || gidJ >= mypn10 ) onnz[leq+k]++;
	      else nnz[leq+k]++;
	    }
	    else {
	      // set scaled identity matrix
	      ierr = prol->SetValues(1,&rowi,1,&colj,&val,INSERT_VALUES);
	      CHKERRQ(ierr);  
	    }
	  }
	}
      }
      leq += fndf;
    }
    assert(leq == nnleq); 
  }

  // assemble (collective on comm_world)
  if( prol != NULL ){
    ierr = prol->Assembly(); CHKERRQ(ierr);
    // clear DAP
    if( getMGType() == PROM_AGGMG && clearDAP ) {
      const PromAggGrid *amggrid = dynamic_cast<const PromAggGrid*>(prevG);
      if( amggrid->amg_smooths_ > 0 ) {
	for( int xx = 0 ; xx < prevG->nNodes_ ; xx++ ) {
	  PromNode *finei = prevG->nodes_[xx];
	  if( finei->nParents() > 0 ) {
	    PromParent **parI = finei->parents_->data_; 
	    for( int II = 0, np = finei->nParents() ; II < np ; II++ ){
	      shpfloat *dap = parI[II]->DAP_;
	      if( dap != NULL ){
#if defined(PROM_DELETE_DAP) 
#error this causes a bug for 2nd+ solves!!!
		PetscFree( dap ); parI[II]->DAP_ = NULL;
#else
		PetscMemzero( dap, parI[II]->getPsz()*sizeof(shpfloat) );
#endif
	      }
	    }
	  }
	}
      }
      if( prevG->gid_extraParents_ != NULL ) {
	TablePos tpos = prevG->gid_extraParents_->getHeadPosition(); 
	while( tpos ) {
	  PromParentArr_extra *pararr = prevG->gid_extraParents_->getNext( tpos );
	  for( int jj = 0, np = pararr->getSize() ; jj < np ; jj++ ) {
	    PromParent *par = pararr->data_[jj];
	    shpfloat *dap = par->DAP_;
	    if( dap != NULL ){
#if defined(PROM_DELETE_DAP)
	      PetscFree( dap ); par->DAP_ = NULL;
#else
	      PetscMemzero( dap, par->getPsz()*sizeof(shpfloat) );
#endif
	    }
	  }
	}
      }
    }
  }

  PetscFunctionReturn(0);
}

//#define USE_MATID
#undef __FUNCT__  
#define __FUNCT__ "PromGrid::GetGhostData" 
/* int PromGrid::GetGhostData() ****************************************
 *
 *  - set up ghost node data: make dependants lists in boundary nodes
 *    called just after PromGrid::SetIndex, and in PromGrid::Repartition.
 *  - ghosts have lid, realproc, realgid
 *  - PromGrid::ghosts have ghosts
 *
 *  INPUT:
 *   fulldata: 
 *
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::GetGhostData( const bool fulldata /*=true*/ )
{
  int            locid, ii, tt, ndeps, pp, np, xx, gid, proc, proc2, ierr;
  const int      gid0 = my0Nd(); //proc_gnode_[comm_.mype()]; 
  PromCoord      crd; xx = sizeof(gfloat); tt = sizeof(MPI_Request);
  const int      reqsz = (tt/xx + ((tt%xx == 0) ? 0 : 1)) * (xx/sizeof(int));
  PromNode      *curr, *n1; const int tag = getNewTag(), tag1 = getNewTag();
  PromGlobalID  *dep;     
  MPI_Status     status; 
  TablePos       tpos; 
  PromList<PromNode*>   *list;
  PromList<MPI_Request*> smessages;                    
  PromTable<PromList<PromNode*>*>  proc_list(57);
  Position       pos;
  
  // collect ghost data in lists
  for( xx = nLocalNd_ ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx]; assert(curr->getState() != PROM_NOTUSED );
    gid = curr->index_.global(); assert(curr->symbData_!=NULL); 
    proc = curr->symbData_->ghostID.proc(ierr); CHKERRQ(ierr);
    if( (list=proc_list.find(proc+1)) == NULL ) {
      list = new PromList<PromNode*>(100); CHKNEW(list,1);
      proc_list.add( list, proc+1 );
    }
    list->addTail( curr ); 
  }
  // isend localid (empty lists, but keep them)
  tpos = proc_list.getHeadPosition(); 
  while( tpos ) {
    list = proc_list.getNext( tpos, &proc ); proc--;
    ii = (2*list->getCount() + 1 + reqsz)*sizeof(int);
    MPI_Request *sreq; ierr = PetscMalloc(ii,&sreq);    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    assert((size_t)pb%sizeof(gfloat) == 0);
    while( !list->isEmpty() ) {
      curr = list->removeHead(); 
      *pb++ = curr->index_.global(); *pb++ = curr->index_();
    }
    *pb++ = -2;
    // send(tag) *<gid, glid>, -2>
    MPI_Isend( buff, pb-buff, MPI_INT, proc/factor_, tag, gComm_, sreq );
  }

  // count bound
  int rbuf[PROM_BSZ]; 
  for( xx = ndeps = 0 ; xx < nLocalNd_ ; xx++ ) { 
    curr = nodes_[xx];
    TablePos tpos = curr->symbData_->adjac.getHeadPosition(); np = 0;
    if(tpos==0 && options_.verbose_>2)
      PetscPrintf(PETSC_COMM_SELF,
		  "[%d]%s WARNING gid %d (%d/%d) has no adjac\n",
		  comm_.mype(),__FUNCT__,curr->index_.global(ierr),
		  xx+1,nLocalNd_);
    while( tpos ) {
      n1 = curr->symbData_->adjac.getNext(tpos); 
      if( n1->isGhost() ){
	proc = n1->symbData_->ghostID.proc(ierr); CHKERRQ(ierr);
	for( pp = 0 ; pp < np ; pp++ ){ if( rbuf[pp] == proc ) break; }
	if( pp == np ) rbuf[np++] = proc;  // new proc
      }
    }
    if( np ) {
      assert( curr->symbData_->dependents == NULL ); // vergin mesh
      curr->symbData_->dependents = new PromList<PromGlobalID*>(np); 
      CHKNEW( curr->symbData_->dependents, 1 );
      ndeps += np; // number of deps
    }
  }
  // recv(tag) bound deps, add bounds to lists
  while( ndeps ){
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE*factor_;
    list = proc_list.find(proc+1); 
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag) *<gid, glid>, -2
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status );
    tt = tt/2; ndeps -= tt;
    while( tt-- ) {
      gid = *pb++; locid = *pb++; xx = gid - gid0; assert(xx>=0&&xx<nLocalNd_);
      curr = nodes_[xx]; 
      list->addTail( curr ); // keep bound nodes for send
      dep = new PromGlobalID( locid, proc ); CHKNEW(dep,1);
      curr->symbData_->dependents->addTail( dep );
    }
    assert(*pb++ == -2);
    PetscFree( buff );
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe(MPI_ANY_SOURCE,tag,gComm_,&ii,&status);  assert(!ii);

  // send bound data, empty lists
  const int gisz = sizeof(gfloat)/sizeof(int); assert(reqsz%gisz == 0);
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    if( !list->isEmpty() ) {
      pos = list->getHeadPosition(); assert(pos!=NULL); tt = 0; 
      while( pos ) {
	curr = list->getNext( pos );
	tt += 1; // gid
	if( fulldata ) {
	  tt += 1; // geq
	  if( tt%gisz != 0 ) tt++;
	  curr->GetWrite_size( np );
	  tt += np + 1; // node + sentinal
	  TablePos tpos = curr->symbData_->adjac.getHeadPosition();
	  assert(tpos!=0); 
	  while( tpos ) {
	    n1 = curr->symbData_->adjac.getNext( tpos ); 
	    if( n1->isBound() || n1->index_() >= nLocalNd_ ) tt++;
	  }
	}
	assert( curr->symbData_->dependents != NULL );
	tt += 1 + 2*curr->symbData_->dependents->getCount(); // ndeps + deps
      }
      int sz = reqsz + tt;
      MPI_Request *sreq; 
      ierr = PetscMalloc(sz*sizeof(int),&sreq); CHKERRQ(ierr);
      smessages.addTail( sreq );
      int *buff = (int*)sreq + reqsz, *pb = buff; 
      assert((size_t)pb%sizeof(gfloat)==0); 
      while( !list->isEmpty() ) {
	curr = list->removeHead(); 
	*pb++ = curr->index_.global(); 
	if( fulldata ) {
	  *pb++ = curr->geq();  
	  if( (size_t)pb%sizeof(gfloat) != 0 ) *pb++ = -2; 
	  curr->Write( &pb );
	  TablePos tpos = curr->symbData_->adjac.getHeadPosition(); 
	  while( tpos ) {
	    n1 = curr->symbData_->adjac.getNext( tpos ); 
	    if( n1->isBound() || n1->index_() >= nLocalNd_ ) {
	      PromContext::Write( n1->index_.global(), pb ); 
	    } 
	  }
	  *pb++ = -1; // sentinal
	}
	// ndeps
	PromList<PromGlobalID*> *dlist = curr->symbData_->dependents;
	assert(dlist!=NULL);
	PromContext::Write( dlist->getCount(), pb ); 
	pos = dlist->getHeadPosition();       assert(pos!=NULL); 
	while( pos ) {
	  dep = dlist->getNext( pos );
	  PromContext::Write( dep->local(), pb ); 
	  PromContext::Write( dep->proc(), pb ); 
	}
      }
      assert(pb-buff == tt);
      //send(tag1)*<gid,geq,[[-1],crd,gid,nd,*<gid>,-1]|ndeps,ndeps*<local,proc>>
      MPI_Isend( buff, pb-buff, MPI_INT, proc/factor_, tag1, gComm_, sreq );
    }
    delete list; 
  }
  // receive ghost data
  int nghosts = nNodes_ - nLocalNd_, my0 = my0Nd();
  while( nghosts ){
    MPI_Probe( MPI_ANY_SOURCE, tag1, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE*factor_;
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag1) *<gid,geq,[[-1],crd,gid,node,*<gid>,-1]|ndeps,ndeps*<lid,pc>> 
    MPI_Recv( buff, tt, MPI_INT, proc/factor_, tag1, gComm_, &status );
    while( pb-buff < tt ) {
      nghosts--;   assert(nghosts>=0);
      PromContext::Read( gid, pb );
      curr = ghosts_.find( gid+1 ); assert(curr!=NULL);
      if( fulldata ) {
	int geq = *pb++; curr->setGEQ(geq); 
	if( (size_t)pb%sizeof(gfloat) != 0 ){ ii = *pb++; assert(ii==-2); } 
	ii = curr->index_();
	curr->Read( parBank_, &pb ); assert(curr->index_() == ii);
	// adjacs
	while(1) {
	  PromContext::Read( ii, pb ); 
	  if( ii == -1 ) break;
	  locid = ii - my0; 
	  if( locid<0 || locid >= nLocalNd_) n1 = ghosts_.find(ii+1);
	  else n1 = nodes_[locid];
	  if( n1 && curr->symbData_->adjac.add(n1) ) {
	    n1->symbData_->adjac.add(curr);
	  }
	}
      }
      // deps
      PromContext::Read( ndeps, pb ); assert( ndeps > 0 );
      assert( curr->symbData_->dependents == NULL ); // ??? 
      curr->symbData_->dependents = new PromList<PromGlobalID*>(ndeps);
      CHKNEW(curr->symbData_->dependents,1);
      while( ndeps-- ) {
	PromContext::Read( locid, pb); PromContext::Read( proc2, pb );
	if( proc2 == comm_.mype() ) { // point back
	  assert( locid == curr->index_() );
	  proc2 = status.MPI_SOURCE * factor_; 
	  locid = gid - proc_gnode_[proc2]; assert( locid >= 0 );
	  dep=new PromGlobalID(locid, proc2);CHKNEW(dep,1);//ref to real node 
	}
	else {
	  dep=new PromGlobalID(locid, proc2);CHKNEW(dep,1);//ref to other ghost
	}
	curr->symbData_->dependents->addTail( dep );
      }
    }
    assert(pb-buff == tt);        PetscFree(buff);
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  // debug
  MPI_Iprobe(MPI_ANY_SOURCE,tag,gComm_,&ii,&status);  assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE,tag1,gComm_,&ii,&status);  assert(!ii);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::PrepForSolve"
/* PromGrid::PrepForSolve ******************************
 *
 */
int PromGrid::PrepForSolve()
{
  int ierr;

  // set asm blocks
  if( isTop() && !isBase() ){ nBlocks_ = 0; assert(blockISs_==NULL);}
  if( blockISs_ == NULL && nBlocks_ != 0 ) {
    // 'nBlocks_' has global number now
    ierr = MakeASMBlocks(); CHKERRQ(ierr);
  }

  ierr = DeleteSymb(); CHKERRQ(ierr);
  assert( (nBlocks_==0) == (blockISs_==NULL) || isBase() );
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::MakeASMBlocks" 
/* PromGrid::MakeASMBlocks *****************************************
 *
 *   - setup blocks for Schwarz methods
 *
 *   INPUT:
 *
 *   SIDE EFFECTS:
 *    - creates "blockISs"
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromGrid::MakeASMBlocks()
{
  int ierr; 
  assert(blockISs_==NULL); // make blockISs
  
  if( isActive() && nLocalNd_ > 0 ) {
    assert( nBlocks_ > 0 );
    // 
    bool overlap = false; int tt;
    ierr = options_.GetOptionInt("-prometheus_schwarz_overlap",&tt,&overlap);
    CHKERRQ(ierr);
    if( overlap ) {
      if( tt != 1 && tt != 2 ) {
	tt = 0;
	if( options_.verbose_>1  )
	  PetscPrintf( gComm_,
		       "[%d](%d) WARNING overlap %d ignored (1 and 2 only)\n",
		       comm_.mype(), getLevel(), tt );
      }
    }
    else tt = 0;
    // make blocks 
    ierr = MakeNodeBlocks( tt );  
    CHKERRQ(ierr);
  }
  else{
    nBlocks_ = 1;
    ierr = PetscMalloc( sizeof(PromIS),&blockISs_); CHKERRQ(ierr); 
    ierr = PromISCreateGeneral(PETSC_COMM_SELF,0,NULL, blockISs_);
    CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::MakeNodeBlocks"   
/*  PromGrid::MakeNodeBlocks *****************************************
 *
 *   - set blocks: nBlocks for node based partitioning
 *
 *   INPUT:
 *    - overlap: 
 *
 *   SIDE EFFECTS:
 *    - sets 'nBlocks_' and 'blockISs_'
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromGrid::MakeNodeBlocks( const int overlap /*=0*/)
{ 
  if( nLocalNd_ == 0 ) return 0;   assert(isActive());
  TablePos tpos;
  int ierr,tt,xx,bi,*pnode_part,my0 = my0Nd(), sbuf[PROM_BSZ];
  const int fact = stiffness_->isComplex() ? 2 : 1;
  PromNode *curr;
  
  // make partitions
  assert(nLocalNd_>0);
  ierr = PetscMalloc(nLocalNd_*sizeof(int),&pnode_part);   CHKERRQ(ierr);
  //
  ierr = PartitionBlocksNodes( nBlocks_, pnode_part );
  CHKERRQ(ierr);
  const int nblocs = nBlocks_;
  //if( special_gids && !special_gids->isEmpty() ) nBlocks_++;
  if( options_.verbose_>1  )
    PetscPrintf(gComm_,
		"[%d](%d)Overlapped ASM, %d blocks (neq=%d), %d overlap\n",
		comm_.mype(), getLevel(), nblocs, nlocalEq(), overlap );

  // accumuate blocks nodes
  assert(nBlocks_>0);
  ierr = PetscMalloc(nBlocks_*sizeof(PromIS),&blockISs_); CHKERRQ(ierr);
  PromList<int> *blocklists = new PromList<int>[nblocs];
  CHKNEW(blocklists,nblocs);
  for( xx = 0; xx < nLocalNd_ ; xx++ ) { // local nodes only
    const int part = pnode_part[xx], gid = nodes_[xx]->index_.global();
    //if( !special_gids || !special_gids->findi(gid+1) ) {
    blocklists[part].addTail(gid+1);
    //}
  }

  // make block IS's
  PromTable<int> table1(227), table2(227);
  for( bi = 0 ; bi < nblocs ; bi++ ) {
    PromList<int> * const blist = &blocklists[bi];
    // increase overlap
    if( overlap > 0 ) {
      Position pos = blist->getHeadPosition(); 
      while( pos ) {
	int gid = blist->getNext( pos ) - 1, xx = gid - my0; 
	table2.add( gid + 1 );
	curr = nodes_[xx];                   assert(xx >= 0 && xx < nLocalNd_);
	PromTable<PromNode*> *adj = &curr->symbData_->adjac; 
	tpos = adj->getHeadPosition();
	while( tpos ) {
	  bool addit = false;
	  PromNode *n1 = adj->getNext( tpos );
	  //if( special_gids && special_gids->findi( gid + 1 ) ) continue; 
	  if( overlap == 2 ) addit = true;
	  else if( n1->isGhost() ) {
	    int p2=n1->symbData_->ghostID.proc(), ev=((p2+comm_.mype())%2==0);
	    if( ev == 0 && p2 < comm_.mype() ) addit = true;
	    if( ev == 1 && p2 > comm_.mype() ) addit = true;
	  }
	  else if( !n1->isGhost() ) {
	    if( pnode_part[n1->index_()] < bi ) addit = true;
	  }
	  if( addit ) table1.add( n1->index_.global()+1 );	  
	}
      }
      while( !table1.isEmpty() ) {
	int gid1 = table1.removeHead();
	if( !table2.findi( gid1 ) ) blist->addTail( gid1 );
      }
      table2.removeAll();
    }
    // count size
    Position pos = blist->getHeadPosition(); tt = 0;
    while( pos ) {
      int gid = blist->getNext( pos ) - 1, lid = gid - my0; 
      if( lid >= 0 && lid < nLocalNd_ ) curr = nodes_[lid];
      else curr = ghosts_.find(gid+1); assert(curr!=NULL);
      tt += curr->getNDF();
    } 
    int *ipr; tt *= fact;
    if( tt > PROM_BSZ ){
      ierr = PetscMalloc( sizeof(int)*tt, &ipr ); CHKERRQ(ierr);
    }
    else ipr = sbuf;
    // create IS
    tt = 0;
    while( !blist->isEmpty() ) {
      int gid = blist->removeHead() - 1, lid = gid - my0;
      if( lid >= 0 && lid < nLocalNd_ ) curr = nodes_[lid];
      else curr = ghosts_.find(gid+1); assert(curr!=NULL);
      int eq = curr->geq()*fact, ndf = curr->getNDF()*fact;
      for( int j=0 ; j < ndf ; j++, tt++, eq++ ) ipr[tt] = eq;
    }
    qsort( (void*)ipr, tt, sizeof(int), prom_intcompar );
    // create 
    ierr = PromISCreateGeneral(PETSC_COMM_SELF, tt, ipr, &blockISs_[bi]);
    CHKERRQ(ierr);
    if( ipr != sbuf ) PetscFree(ipr);
  } 
  PetscFree(pnode_part);                                //CHKERRQ(ierr);
  delete [] blocklists;

  // create special (provided) block
  /*if( special_gids && !special_gids->isEmpty() ) {
    if( options_.verbose_> 2 ){
    PetscPrintf( (options_.verbose_>1) ? PETSC_COMM_SELF : PETSC_COMM_WORLD,
    "\t[%d] Set 'special' ASM domains with %d nodes\n",
    comm_.mype(),special_gids->getCount());	 
    }
    // count size
    tpos = special_gids->getHeadPosition(); tt = 0; 
    while( tpos ) {
    int gid = special_gids->getNext(tpos) - 1, lid = gid - my0;
    if( lid >= 0 && lid < nLocalNd_ ) curr = nodes_[lid];
    else {
    curr = ghosts_.find(gid+1); 
    if( options_.verbose_ > 2  ){
    PetscPrintf((options_.verbose_>1)?PETSC_COMM_SELF:PETSC_COMM_WORLD,
    "\t\t[%d] Warning 'special' ASM domain has ghost %d\n",
    comm_.mype(),gid);	 
    }
    }
    if( curr != NULL ) tt += curr->getNDF();
    else {
    //int kk = gid_ndf_geq->find( gid + 1 ); // has 'factor'
    //if( kk == 0 ) 
    SETERRQ1(1,"pen CR not found %d",gid);
    //int ndf = kk % PROM_MAXNDF; 
    //tt += ndf;
    }
    }
    int *ipr;
    if( tt > PROM_BSZ ){
    ierr = PetscMalloc( sizeof(int)*tt, &ipr ); CHKERRQ(ierr);
    }
    else ipr = sbuf;
    // create
    tpos = special_gids->getHeadPosition(); tt = 0; 
    while( tpos ) {
    int gid = special_gids->getNext(tpos) - 1, lid = gid - my0, eq, ndf;
    if( lid >= 0 && lid < nLocalNd_ ) curr = nodes_[lid];
    else curr = ghosts_.find(gid+1); 
    if( curr != NULL ) { eq = curr->geq()*fact; ndf = curr->getNDF()*fact; }
    else {
    //int kk = gid_ndf_geq->find( gid + 1 );
    //if( kk == 0 ) 
    SETERRQ1(1,"pen CR not found %d",gid);
    //ndf = kk % PROM_MAXNDF; eq = kk / PROM_MAXNDF;
    }
    for( int j=0 ; j < ndf ; j++, tt++, eq++ ) ipr[tt] = eq;
    }
    // 
    qsort( (void*)ipr, tt, sizeof(int), prom_intcompar );
    ierr = PromISCreateGeneral(PETSC_COMM_SELF,tt,ipr,&blockISs_[nBlocks_-1]);
    CHKERRQ(ierr);
    if( ipr != sbuf ) PetscFree(ipr);
    }*/
  
  return 0;
}

static int get_norm( const int indf, const int jndf, 
		     const double pAij[], double *ret ) 
{
  int j,i, bs2 = indf*jndf;
  double sum, mx, mx2; const double *aa;
  
  for( i = 0, mx = 0., aa = pAij ; i < jndf ; i++ ){
    for( j = 0, sum = 0. ; j < indf; j++, aa++ ) sum += fabs(*aa);
    if( sum > mx ) mx = sum;
  }
  for( i = 0, mx2 = 0., aa = pAij ; i < indf ; i++, pAij++ ){
    for( j = 0, sum = 0.; j < bs2 ; j += indf ) sum += fabs(pAij[j]);
    if( sum > mx2 ) mx2 = sum;
  } 
  *ret = mx + mx2; 
  assert(*ret >= 0.0);
    
  return 0;
} 

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::PartitionBlocksNodes" 
/*  PromGrid::PartitionBlocksNodes ****************************************
 *
 *   - do local partition
 *
 *   INPUT:
 *    - pnode_part[nLocalNd_]: local nodes to block map (out)
 *
 *   SIDE EFFECTS:
 *    - sets 'nBlocks_' and 'blockISs_'
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromGrid::PartitionBlocksNodes( const int, int *pnode_part ) 
{
  int ii,mype=comm_.mype();const int nloc=nlocal();
  PromNode *curr; 
  
  if( nBlocks_ == 1 || nloc < 1 ){
    for( ii = 0 ; ii < nloc ; ii++ ) pnode_part[ii] = 0;
    return 0;
  }
  else if( nBlocks_ > nloc / 2 ) {
    PetscPrintf( PETSC_COMM_SELF, "[%d]%s WARNING reseting nBlocks=%d\n",
		 mype, __FUNCT__,nBlocks_);
    nBlocks_ = nloc;
    for( ii = 0 ; ii < nloc ; ii++ ) pnode_part[ii] = ii;
    return 0;
  }
#if defined(PROM_HAVE_METIS)
  { // set 'xadj' & 'adjncy' adjacentcy data
    int *egwts=NULL, locid, ierr, lid; 
    PromNode *n1; const int fact = stiffness_->isComplex() ? 2 : 1;
    double *diag = NULL, Aii; idxtype *adjncy, *xadj; 
    const int my0 = my0Nd(), next0 = my0 + nlocal();
    PromMatrix *AA = stiffness_; assert(AA!=NULL);
    TablePos tpos;

    ierr = PetscMalloc((nloc+1)*sizeof(idxtype),&xadj); CHKERRQ(ierr);    
    if( AA != NULL ) {
      ierr = PetscMalloc((nloc+1)*sizeof(double),&diag); CHKERRQ(ierr);
    }
    for( locid = ii = 0 ; locid < nloc ; locid++ ) {
      curr = nodes_[locid]; const int ndf = curr->getNDF()*fact;
      //if( gids && gids->findi(curr->index_.global() + 1)) continue; 
      if( AA != NULL ) {
	int adjacs[PROM_MAXNIEGH],ncols,gid; double *Ai[PROM_MAXNIEGH], tv;
	ierr = AA->GetLocalNodeRow(locid,PROM_MAXNIEGH,ncols,adjacs,Ai); 
	CHKERRQ(ierr);
	if( ncols == 0 ) {
	  PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s WARNING:have zero row\n",mype,__FUNCT__);
	  diag[locid] = 1.0;
	}
	else {
	  diag[locid] = 0.0;
	  for( int nn = 0 ; nn < ncols ; nn++ ) {
	    gid = adjacs[nn];
	    //if( gids && gids->findi( gid + 1 ) ) continue; 
	    if( (lid=gid-my0) >= 0 && gid < next0 ) {
	      if( lid == locid ) {
		get_norm( ndf, ndf, Ai[nn], &tv );
		diag[locid] = tv;
	      }
	      else{
		if( curr->symbData_->adjac.find( nodes_[lid] ) ) ii++;
	      }
	    }
	  }
	  assert( diag[locid] > 0.0 );
	}
      }
      else {
	PromTable<PromNode*> *adj = &curr->symbData_->adjac; 
	assert(curr->symbData_ != NULL);
	tpos = adj->getHeadPosition(); 
	while( tpos ) {
	  n1 = adj->getNext(tpos); 
	  //if( gids && gids->findi( n1->index_.global() + 1 ) ) continue; 
	  if( !n1->isGhost() && n1 != curr ) ii++;
	}
      }
    }
    const int numadjac = ii; assert( ii%2 == 0 ); assert( numadjac > 0 );
    ierr = PetscMalloc( numadjac * sizeof(idxtype),&adjncy ); CHKERRQ(ierr);
    if( AA != NULL ) {
      ierr = PetscMalloc( numadjac * sizeof(idxtype),&egwts ); CHKERRQ(ierr);
    }
    for( locid = ii = 0 ; locid < nloc ; locid++ ){ //NB:iter. same as egwts!
      curr = nodes_[locid];  xadj[locid] = ii; const int ndf=curr->getNDF()*fact;
      //if( gids && gids->findi( curr->index_.global() + 1 ) ) continue; 
      if( AA != NULL ) {
	int adjacs[PROM_MAXNIEGH],ncols,gid;
	double *Ai[PROM_MAXNIEGH],Aij,temp, tv;
	Aii = diag[locid];
	ierr = AA->GetLocalNodeRow(locid,PROM_MAXNIEGH,ncols,adjacs,Ai); 
	CHKERRQ(ierr);
	for( int nn = 0 ; nn < ncols ; nn++ ) {
	  gid = adjacs[nn];
	  //if( gids && gids->findi( gid + 1 ) ) continue; 
	  if( (lid=gid-my0) >= 0 && gid < next0 && lid != locid ){ 
	    if( curr->symbData_->adjac.find( nodes_[lid] ) ) {
	      get_norm( ndf, nodes_[lid]->getNDF()*fact, Ai[nn], &tv );
	      Aij = tv;
	      temp = Aij / sqrt( Aii * diag[lid] );
	      egwts[ii] = (int)(temp * 10.0) + 1;
	      adjncy[ii++] = lid;
	    }
	  }
	}
      }
      else {
	PromTable<PromNode*> *adj = &curr->symbData_->adjac; 
	tpos = adj->getHeadPosition(); 
	while( tpos ) {
	  n1 = adj->getNext(tpos);
	  //if( gids && gids->findi( n1->index_.global() + 1 ) ) continue; 
	  if( !n1->isGhost() && n1 != curr ) adjncy[ii++] = n1->index_();
	}
      }
    }
    if(numadjac!=ii)SETERRQ3(1,"numadjac(%d)!=ii(%d) nloc=%d",numadjac,ii,nloc);
    xadj[nloc] = ii; 

    // get local partition
    int options[5]={0,3,1,1,0},ncon=1, weightflag = 0;
    float ubvec = 1.05, *tpwgts; MPI_Comm scomm = MPI_COMM_SELF;
    ierr = PetscMalloc(nBlocks_*sizeof(float), &tpwgts); CHKERRQ(ierr);
    for( ii = 0 ; ii < nBlocks_ ; ii++ ) tpwgts[ii] = 1./(float)nBlocks_;
    if( egwts != NULL ) weightflag++;
    int nmbng = 0, edgecut = -1, np = nBlocks_, nn = nloc;
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_METIS] );
    ParMETIS_V3_PartKway( &nn, xadj, adjncy, NULL, egwts, &weightflag, &nmbng, &ncon, &np, tpwgts, &ubvec, options, &edgecut, pnode_part, &scomm );
    perf_mon_.EventEnd( PromContext::prom_logs_[LOG_METIS] );
    // free buffers
    PetscFree(xadj);  PetscFree(adjncy); PetscFree(tpwgts);
    if( AA != NULL ) {
      PetscFree(egwts);  PetscFree(diag);  
    }
  }
#else
  if( nBlocks_ > nloc/3 ) {
    for( ii = 0 ; ii < nloc ; ii++ ) pnode_part[ii] = ii/2;
    nBlocks_ = nloc/2;
  }
  else {
    int xx, rooti, ndone, nb, block, targnb = nloc/nBlocks_;
    PromTable<int> table; PromList<int> rootlist;
    // breath first search -- outputs 'nbklks' 
    rooti = 0; 
    rootlist.addTail( rooti + 1 );
    table.add( nodes_[rooti]->index_() + 1 );
    pnode_part[rooti] = 0;            assert(nodes_[rooti]->symbData_ != NULL);
    for( ndone = 1, block = 0 ; ndone < nloc ; ) {
      for( nb = 0 ; nb < targnb && ndone < nloc ; ) {
	assert(!rootlist.isEmpty());
	rooti = rootlist.removeHead() - 1; assert(rooti>=0 && rooti<nloc);
	curr = nodes_[rooti];
	TablePos tpos = curr->symbData_->adjac.getHeadPosition();
	while( tpos ) {
	  PromNode *n1 = curr->symbData_->adjac.getNext( tpos );
	  if( !n1->isGhost() && table.add( n1->index_() + 1 ) ) {
	    ii = n1->index_();
	    rootlist.addTail( ii + 1 );
	    pnode_part[ii] = block;
	    nb++; ndone++;
	  }
	}
	if( ndone < nloc && rootlist.isEmpty() ) {
	  for( xx = 0 ; xx < nloc ; xx++ ) {
	    if( table.add( xx + 1 ) ) {
	      rootlist.addTail( xx + 1 );
	      pnode_part[xx] = block; 
	      nb++; ndone++;
	      break;
	    }
	  }
	}
      }
      block++;
    }
    nBlocks_ = block; // (out)
  }
#endif
  return 0;  
}

/***********************************************************************
 *                           PromCRMatrix                              *
 ***********************************************************************/
#undef __FUNCT__  
#define __FUNCT__ "PromCRMatrix::PromCRMatrix"
/* PromCRMatrix::PromCRMatrix *********************************************
 *
 */ 
PromCRMatrix::PromCRMatrix(const PromGrid*grid,int nA,int*pnA,int nB,int*pnB):  
  PromMatrix( grid, nA, pnA, nB, pnB ), D_(NULL), C_(NULL), 
  Ct_(NULL), CDinvCt_(NULL)
{
}

/***********************************************************************
 *                           PromComplexMatrix                         *
 ***********************************************************************/

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::PromComplexMatrix"
/* PromComplexMatrix::PromComplexMatrix ******************************
 *
 */ 
//! constructor
PromComplexMatrix::PromComplexMatrix( const PromGrid *grid, int nA,int *pnA,
				      int nB, int *pnB ) :
  PromMatrix( grid, nA, pnA, nB, pnB, 2 )
{
#if defined(PETSC_USE_BOPT_g)
  PetscPrintf(PETSC_COMM_WORLD,"[%d]%s K1 Eq. R. Form %d\n",0,__FUNCT__,map_.nglobalEq());
#endif
}
//! constructor
PromComplexMatrix::PromComplexMatrix(const PromMap &map, int nA,int *pnA,
				     int nB,int*pnB,int lcols/*=-1*/,
				     int gcols/*=-1*/):
  PromMatrix( map, nA, pnA, nB, pnB, lcols, gcols, 2 )
{
#if defined(PETSC_USE_BOPT_g)
  PetscPrintf(PETSC_COMM_WORLD,"[%d]%s K1 Eq. R. Form %d\n",0,__FUNCT__,map_.nglobalEq());
#endif
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::GetOwnershipRange"
/* PromComplexMatrix::GetOwnershipRange ******************************
 *
 */ 
int PromComplexMatrix::GetOwnershipRange(int *first, int *end)const
{
  assert(this!=NULL);
  int f,e,ierr = MatGetOwnershipRange( mat_, &f, &e ); CHKERRQ(ierr);
  *first = f/2; *end = e/2; assert(e%2==0);
  return 0; 
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::GetLocalNodeRow"
/* PromComplexMatrix::GetLocalNodeRow *************************************
 * 
 *  get REAL values - condense native data
 */ 
int PromComplexMatrix::GetLocalNodeRow(const int brow, const int len, 
				       int &ncols, int adjacs[], 
				       double **Aij /*=NULL*/ ) const
{
  assert(this!=NULL);
  // only accesses half of data ???
  int ierr = PromMatrix::GetLocalNodeRow(brow, len, ncols, adjacs, Aij);
  CHKERRQ(ierr);
  
  return 0; 
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::GetRow"
/* PromComplexMatrix::GetRow_r *************************************
 *
 *  get REAL values - condense native data
 */ 
int PromComplexMatrix::GetRow(int row, int *ncols,int **colpp,
			      double **arrdat)const
{
  assert(this!=NULL);
  // only accesses half of data
  int ierr = PromMatrix::GetRow(2*row, ncols, colpp, arrdat);
  CHKERRQ(ierr);
  const int num = *ncols/2; assert( (*ncols)%2==0 );
  if( colpp != NULL ) {
    int *arr = *colpp; 
    for(int ii=0,jj=0 ; jj<num ; jj++, ii+=2) arr[jj] = arr[ii]/2;
  }
  if( arrdat != NULL ) {
    double *arr = *arrdat; 
    for(int ii=0,jj=0 ; jj<num ; jj++, ii+=2) arr[jj] = arr[ii];
  }
  *ncols = num;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::GetRow_i"
/* PromComplexMatrix::GetRow_i *************************************
 *
 *  get IMAGINARY values - condense native data
 */ 
int PromComplexMatrix::GetRow_i(int row, int *ncols,int **colpp,
				double **arrdat)const
{
  // only accesses half of data (and neg sign on imaginary part!
  int ierr = PromMatrix::GetRow(2*row+1, ncols, colpp, arrdat);
  CHKERRQ(ierr);
  const int num = *ncols/2;
  if( colpp != NULL ) {
    int *arr = *colpp; 
    for(int ii=0,jj=0 ; jj<num ; jj++, ii+=2) arr[jj] = arr[ii]/2;
  }
  if( arrdat != NULL ) {
    double *arr = *arrdat; 
    for(int ii=0,jj=0 ; jj<num ; jj++, ii+=2) arr[jj] = arr[ii];
  }
  *ncols = num;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::RestoreRow"
/* PromComplexMatrix::RestoreRow**********************************
 *
 */ 
int PromComplexMatrix::RestoreRow(int row,int*ncols,int**colpp/*=NULL*/,
				  double**arrdat/*=NULL*/)const
{
  return PromMatrix::RestoreRow(row,ncols,colpp,arrdat);
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::SetValues"
/* PromComplexMatrix::SetValues *************************************
 *
 */ 
int PromComplexMatrix::SetValues(const int nrow, int *pgeqi, 
				 const int ncol, int *colp, double *vals, 
				 const int add_type )
{
  int ii,ierr,bufferi[PROM_BSZ],bufferj[PROM_BSZ];

  if( nrow > PROM_BSZ || ncol > PROM_BSZ ) {
    for(ii=0 ; ii<nrow ; ii++) pgeqi[ii] *= 2; // upper left corner
    for(ii=0 ; ii<ncol ; ii++) colp[ii] *= 2;
    ierr = PromMatrix::SetValues(nrow,pgeqi,ncol,colp,vals,add_type);
    CHKERRQ(ierr);
    for(ii=0 ; ii<nrow ; ii++) pgeqi[ii] += 1; // lower right corner
    for(ii=0 ; ii<ncol ; ii++) colp[ii] += 1; 
    ierr = PromMatrix::SetValues(nrow,pgeqi,ncol,colp,vals,add_type);
    CHKERRQ(ierr);
    for(ii=0 ; ii<nrow ; ii++) pgeqi[ii] = (pgeqi[ii]-1)/2;
    for(ii=0 ; ii<ncol ; ii++) colp[ii] = (colp[ii]-1)/2;
  }
  else {
    for(ii=0 ; ii<nrow ; ii++) bufferi[ii] = 2*pgeqi[ii];
    for(ii=0 ; ii<ncol ; ii++) bufferj[ii] = 2*colp[ii];
    ierr = PromMatrix::SetValues(nrow,bufferi,ncol,bufferj,vals,add_type);
    CHKERRQ(ierr);
    for(ii=0 ; ii<nrow ; ii++) bufferi[ii] += 1;
    for(ii=0 ; ii<ncol ; ii++) bufferj[ii] += 1;
    ierr = PromMatrix::SetValues(nrow,bufferi,ncol,bufferj,vals,add_type);
    CHKERRQ(ierr);
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix:: SetValues_i"
/* PromComplexMatrix::SetValues_i ***********************************
 *
 */ 
int PromComplexMatrix::SetValues_i(const int nrow, int *pgeqi, 
				   const int ncol, int *colp, 
				   double *vals, const int add_type )
{
  int ii,ierr,bufferi[PROM_BSZ],bufferj[PROM_BSZ];
  const int len=nrow*ncol;

  if( nrow > PROM_BSZ || ncol > PROM_BSZ ) {
    for(ii=0 ; ii<nrow ; ii++) pgeqi[ii] = 2*pgeqi[ii] + 1;
    for(ii=0 ; ii<ncol ; ii++) colp[ii] = 2*colp[ii]; // lower left corner
    ierr = PromMatrix::SetValues(nrow,pgeqi,ncol,colp,vals,add_type);
    CHKERRQ(ierr);
    for(ii=0 ; ii<nrow ; ii++) pgeqi[ii] -= 1;        // upper right corner
    for(ii=0 ; ii<ncol ; ii++) colp[ii] += 1;
    for(ii=0 ; ii<len ; ii++) vals[ii] *= -1.0;       // form 1
    ierr = PromMatrix::SetValues(nrow,pgeqi,ncol,colp,vals,add_type);
    CHKERRQ(ierr);
    for(ii=0 ; ii<nrow ; ii++) pgeqi[ii] = pgeqi[ii]/2;
    for(ii=0 ; ii<ncol ; ii++) colp[ii] = (colp[ii]-1)/2;
  }
  else {
    for(ii=0 ; ii<nrow ; ii++) bufferi[ii] = 2*pgeqi[ii] + 1;
    for(ii=0 ; ii<ncol ; ii++) bufferj[ii] = 2*colp[ii];
    ierr = PromMatrix::SetValues(nrow,bufferi,ncol,bufferj,vals,add_type);
    CHKERRQ(ierr);
    for(ii=0 ; ii<nrow ; ii++) bufferi[ii] -= 1;
    for(ii=0 ; ii<ncol ; ii++) bufferj[ii] += 1;
    for(ii=0 ; ii<len ; ii++) vals[ii] *= -1.0; // form 1
    ierr = PromMatrix::SetValues(nrow,bufferi,ncol,bufferj,vals,add_type);
    CHKERRQ(ierr);
  }
  for(ii=0 ; ii<len ; ii++) vals[ii] *= -1.0; // form 1
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::SetValuesBlocked_i"
/* PromComplexMatrix::SetValuesBlocked_i ***************************
 *
 */ 
int PromComplexMatrix::SetValuesBlocked_i(const int nrow, int *prow, 
					  const int ncol,int *pcol, 
					  double *vals, const int add_t)
{
  int ierr;
  ierr = SetValuesBl_private(nrow, prow, ncol, pcol, vals, add_t, true);
  CHKERRQ(ierr);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::SetValuesBlocked"
/* PromComplexMatrix::SetValuesBlocked ****************************
 *
 * - real part assembly
 */ 
int PromComplexMatrix::SetValuesBlocked(const int nrow, int *prow, 
					const int ncol, int *pcol,
					double *vals, const int add_t)
{
  int ierr;
  ierr = SetValuesBl_private(nrow, prow, ncol, pcol, vals, add_t, false);
  CHKERRQ(ierr);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::ZeroRows"
/* PromComplexMatrix::ZeroRows *************************************
 *
 */ 
int PromComplexMatrix::ZeroRows( PromIS is, double diags /*=0.0*/)
{
  int tt,ii,jj,ierr,sz,newIds[32768]; 
  const int *ids;

  ierr = PromISGetIndices( is, &ids );               CHKERRQ(ierr);
  ierr = PromISGetSize( is, &sz );                   CHKERRQ(ierr);
  if( 2*sz > 32768 ) SETERRQ1(1,"IS size (%d) too big",sz);
  for(ii=jj=0;ii<sz;ii++,jj+=2) {
    tt = 2*ids[ii]; newIds[jj] = tt; newIds[jj+1] = tt + 1;
  }
  // create/set/destory
  PromIS newIS;
  ierr = PromISCreateGeneral(PETSC_COMM_SELF,2*sz,newIds,&newIS);
  CHKERRQ(ierr);
  ierr = PromMatrix::ZeroRows( newIS, diags ); CHKERRQ(ierr);
  ierr = PromISDestroy( newIS ); CHKERRQ(ierr);

  return 0;
}

/***********************************************************************
 *                           PromComplexVector                         *
 ***********************************************************************/
#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::Norm2"
/* PromComplexVector::Norm2 *****************************
 *
 */
int PromComplexVector::Norm2( double * const norm ) const
{
  int ierr,ii,sz=getn(); 
  double sum,treal,timag,*pv1;
  ierr = GetArray( &pv1 );           CHKERRQ(ierr);
  const double *vs = pv1;
  for(ii=0,sum=0.0;ii<sz;ii++) {
    treal = *vs++; timag = *vs++; 
    sum += (treal*treal + timag*timag); // conjagate
  }
  ierr = RestoreArray( &pv1 );       CHKERRQ(ierr);

  // construct
  MPI_Allreduce( &sum, &treal, 1, MPI_DOUBLE, MPI_SUM, map_.comm_());
  *norm = (double)sqrt( treal );

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::NormMax"
/* PromComplexVector::NormMax ***************************
 *
 */ 
int PromComplexVector::NormMax( double * const norm )const
{
  int ierr,ii,sz=getn(); 
  double temp,maxv,treal,timag,*pv1;
  ierr = GetArray( &pv1 );           CHKERRQ(ierr);
  const double *vs=pv1;
  for(ii=0,maxv=0.0;ii<sz;ii++) {
    treal = *vs++; timag = *vs++;
    temp = sqrt(treal*treal + timag*timag);  // conjagate
    if( temp > maxv ) maxv = temp;
  }
  ierr = RestoreArray( &pv1 );       CHKERRQ(ierr);

  // construct
  MPI_Allreduce(&maxv, &temp, 1, MPI_DOUBLE, MPI_MAX, map_.comm_());
  *norm = (double)temp;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::SetValues"
/* PromComplexVector::SetValues ************************
 *
 */ 
int PromComplexVector::SetValues( const int nrow, int *rowp, 
				  const double *vals,
				  const int add_type )
{
  int ierr,ii,buff[PROM_BSZ];

  if( nrow > PROM_BSZ ) {
    for(ii=0;ii<nrow;ii++)rowp[ii] = 2*rowp[ii];
    ierr = PromVector::SetValues(nrow, rowp, vals, add_type);
    CHKERRQ(ierr);
    for(ii=0;ii<nrow;ii++)rowp[ii] = rowp[ii]/2;
  }
  else {
    for(ii=0;ii<nrow;ii++)buff[ii] = 2*rowp[ii];
    ierr = PromVector::SetValues(nrow, buff, vals, add_type);
    CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::SetValues_i"
/* PromComplexVector::SetValues_i **********************
 *
 */ 
int PromComplexVector::SetValues_i( const int nrow, int *rowp, 
				    const double *vals,
				    const int add_type )
{
  int ierr,ii,buff[PROM_BSZ];
  
  if( nrow > PROM_BSZ ) {
    for(ii=0;ii<nrow;ii++)rowp[ii] = 2*rowp[ii] + 1;
    ierr = PromVector::SetValues(nrow, rowp, vals, add_type);
    CHKERRQ(ierr);
    for(ii=0;ii<nrow;ii++)rowp[ii] = (rowp[ii]-1)/2;
  }
  else {
    for(ii=0;ii<nrow;ii++)buff[ii] = 2*rowp[ii] + 1;
    ierr = PromVector::SetValues(nrow, buff, vals, add_type);
    CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::Sqrt"
/* PromComplexVector::Sqrt ******************************
 *
 */ 
int PromComplexVector::Sqrt()
{
  PetscPrintf(PETSC_COMM_WORLD,"[%d]%s WARNING ...\n",0,__FUNCT__);
  return PromVector::Sqrt();
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::PointwiseMult"
/* PromComplexVector::PointwiseMult ***************************
 *
 */ 
int PromComplexVector::PointwiseMult(const PromVector_base *const X, 
				     const PromVector_base *const Y)
{
  return 1;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::PointwiseDiv"
/* PromComplexVector::PointwiseDiv *****************************
 *
 */ 
int PromComplexVector::PointwiseDiv( const PromVector_base *const X,
				     const PromVector_base *const Y)
{
  return 1;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::Reciprocal"
/* PromComplexVector::Reciprocal ***********************************
 *
 */ 
int PromComplexVector::Reciprocal()
{
  return 1;
}

#if defined(PROM_HAVE_COMPLEX)

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cScale"
/* PromComplexVector::cScale *****************************
 *
 */ 
int PromComplexVector::cScale( PromComplex &a )
{
  int ierr,ii,sz=getn(); double *pv1; 
  ierr = GetArray( &pv1 );           CHKERRQ(ierr);
  double *vs = pv1, alpha_r = a.real(), alpha_i = a.imag(); 
  for(ii=0;ii<sz;ii++) {
    double t2_r = vs[0], t2_i = vs[1];
    *vs++ = (alpha_r*t2_r - alpha_i*t2_i); 
    *vs++ = (alpha_r*t2_i + t2_r*alpha_i);
  }
  ierr = RestoreArray( &pv1 );       CHKERRQ(ierr);
 
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cDot"
/* PromComplexVector::cDot *******************************
 *
 *    val = (x,y) = this^H X,
 */ 
int PromComplexVector::cDot( const PromVector_base*const X, 
			    PromComplex &dot ) const
{
  assert(this!=NULL);
  int ierr,ii,sz=getn();
  double rsum,isum,t1,t2,t1_i,t2_i,*pv,*pvX;
  ierr = GetArray( &pv );           CHKERRQ(ierr);
  ierr = X->GetArray( &pvX );       CHKERRQ(ierr);
  const double *vs=pv,*vsX=pvX;
  for(ii=0,rsum=isum=0.0;ii<sz;ii++){
    t1 = *vs++;  t1_i = *vs++;
    t2 = *vsX++; t2_i = *vsX++;
    rsum += (t1*t2 - t1_i*t2_i); isum += (t1*t2_i + t2*t1_i);
  }
  ierr = RestoreArray( &pv );       CHKERRQ(ierr);
  ierr = X->RestoreArray( &pvX );   CHKERRQ(ierr);
  
  // construct
  double dot2[2] = {rsum, isum}, dd[2];
  MPI_Allreduce(dot2, dd, 2, MPI_DOUBLE, MPI_SUM, map_.comm_());
  dot = PromComplex(dd[0], dd[1]);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cMDot"
/* PromComplexVector::cMDot *************************************
 *
 *  val = (x,y) = YY^H * this,  conjugate transpose 'Y' 
 */ 
int PromComplexVector::cMDot( const int nv, 
			     const PromVector_base*const*YY, 
			     PromComplex xx[] )const
{
  assert(this!=NULL); if(nv>2048)SETERRQ1(1,"nv(%d) > 2048",nv); 
  int ii,ierr,jj,sz=getn(); 
  double t1,t2_r,t1_i,t2_i,*vs,*pv,buff[2*2048];;
  double *pvY[2048],*vv[2048],rsum[2*2048],*isum=rsum+nv;
  
  for(jj=0;jj<nv;jj++){
    ierr = YY[jj]->GetArray( &pvY[jj] ); CHKERRQ(ierr);
    rsum[jj] = isum[jj] = 0.0;
    vv[jj] = pvY[jj];
  }
  ierr = GetArray( &pv );                CHKERRQ(ierr);
  vs = pv;
  //
  for(ii=0;ii<sz;ii++) {
    t1 = *vs++; t1_i = *vs++;
    for(jj=0;jj<nv;jj++){
      t2_r = *vv[jj]++; t2_i = -(*vv[jj]++); // conjugate 
      rsum[jj] += (t1*t2_r - t1_i*t2_i);
      isum[jj] += (t1*t2_i + t2_r*t1_i);
    }
  }
  ierr = RestoreArray( &pv );       CHKERRQ(ierr);
  //
  MPI_Allreduce(rsum, buff, 2*nv, MPI_DOUBLE, MPI_SUM, map_.comm_());
  //
  for(jj=0,pv=buff+nv;jj<nv;jj++){
    ierr = YY[jj]->RestoreArray( &pvY[jj] ); CHKERRQ(ierr);
    xx[jj] = PromComplex(buff[jj], pv[jj]);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cMAXPY"
/* PromComplexVector::cMAXPY ***************************
 *
 * this = this + sum( alpha[j] * x[j] )
 */ 
int PromComplexVector::cMAXPY(const int nv, PromComplex a[],
			      const PromVector_base*const*YY)
{
  assert(this!=NULL); assert(nv<4096); 
  int ierr,jj,ii,sz=getn();
  double alpha_r,t2_r,alpha_i,t2_i,sum_r,sum_i,*pv,*pvY[4096],*vv[4096];
  for(jj=0;jj<nv;jj++){
    ierr = YY[jj]->GetArray( &pvY[jj] ); CHKERRQ(ierr);
    vv[jj] = pvY[jj];
  }
  ierr = GetArray( &pv );           CHKERRQ(ierr);
  double *vs = pv;
  for( ii=0 ; ii<sz ; ii++) {
    for(jj=0,sum_r=sum_i=0.0;jj<nv;jj++){
      alpha_r = a[jj].real(); alpha_i = a[jj].imag(); 
      t2_r = *vv[jj]++; t2_i = *vv[jj]++; 
      sum_r += (alpha_r*t2_r - alpha_i*t2_i); 
      sum_i += (alpha_r*t2_i + t2_r*alpha_i);
    }
    *vs++ += sum_r;
    *vs++ += sum_i;
  }
  ierr = RestoreArray( &pv );       CHKERRQ(ierr);
  for(jj=0;jj<nv;jj++){
    ierr = YY[jj]->RestoreArray( &pvY[jj] ); CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cAXPY"
/* PromComplexVector::cAXPY **************************
 *
 *  this <- this + a * X
 */ 
int PromComplexVector::cAXPY(const PromComplex &a,
			     const PromVector_base *X)
{
  assert(this!=NULL); 
  int ierr,ii,sz=getn(); 
  double alpha_r,t2_r,alpha_i,t2_i,*pv,*Xpv;
  ierr = GetArray( &pv );           CHKERRQ(ierr);
  ierr = X->GetArray( &Xpv );       CHKERRQ(ierr);
  const double *Xvs = Xpv; double *vs = pv;
  alpha_r = a.real(); alpha_i = a.imag(); 
  for( ii=0 ; ii<sz ; ii++) {
    t2_r = *Xvs++; t2_i = *Xvs++;
    *vs++ += (alpha_r*t2_r - alpha_i*t2_i); 
    *vs++ += (alpha_r*t2_i + t2_r*alpha_i);
  }
  ierr = RestoreArray( &pv );       CHKERRQ(ierr);
  ierr = X->RestoreArray( &Xpv );   CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cWAXPY"
/* PromComplexVector::cWAXPY **************************
 *
 *  this(w) <- a * X + Y
 */ 
int PromComplexVector::cWAXPY( const PromComplex &a, const PromVector_base *X,
			       const PromVector_base *Y)
{
  assert(this!=NULL); 
  int ierr,ii,sz=getn(); 
  double alpha_r,t2_r,alpha_i,t2_i,*pv,*Xpv,*Ypv;
  ierr = GetArray( &pv );           CHKERRQ(ierr);
  ierr = X->GetArray( &Xpv );       CHKERRQ(ierr);
  ierr = Y->GetArray( &Ypv );       CHKERRQ(ierr);
  const double *Xvs = Xpv; const double *Yvs = Ypv; double *vs = pv;
  alpha_r = a.real(); alpha_i = a.imag(); 
  for( ii=0 ; ii<sz ; ii++) {
    t2_r = *Xvs++; t2_i = *Xvs++;
    *vs++ = (alpha_r*t2_r - alpha_i*t2_i) + *Yvs++; 
    *vs++ = (alpha_r*t2_i + t2_r*alpha_i) + *Yvs++;
  }
  ierr = RestoreArray( &pv );       CHKERRQ(ierr);
  ierr = X->RestoreArray( &Xpv );   CHKERRQ(ierr);
  ierr = Y->RestoreArray( &Ypv );   CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cAYPX"
/* PromComplexVector::cAYPX **************************
 *
 *  this <- a * this + X
 */ 
int PromComplexVector::cAYPX( const PromComplex &a,
			      const PromVector_base *X)
{
  assert(this!=NULL); 
  int ierr,ii,sz=getn(); 
  double alpha_r,t1_r,t1_i,alpha_i,*pv,*Xpv;

  ierr = GetArray( &pv );           CHKERRQ(ierr);
  ierr = X->GetArray( &Xpv );       CHKERRQ(ierr);
  const double *Xvs = Xpv; double *vs = pv;
  alpha_r = a.real(); alpha_i = a.imag(); 
  for( ii=0 ; ii<sz ; ii++) {
    t1_r = vs[0]; t1_i = vs[1];
    *vs++ = (alpha_r*t1_r - alpha_i*t1_i) + *Xvs++; 
    *vs++ = (alpha_r*t1_i + alpha_i*t1_r) + *Xvs++;
  }
  ierr = RestoreArray( &pv );       CHKERRQ(ierr);
  ierr = X->RestoreArray( &Xpv );   CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromComplexVector::cAXPBY"
/* PromComplexVector::cAXPBY **************************
 *
 *  this = a * X + b * this
 */ 
int PromComplexVector::cAXPBY( PromComplex &a, PromComplex &b, 
			       const PromVector_base *const X ) 
{
  assert(this!=NULL);
  int ierr,ii,sz=getn(); 
  double t1_r,t1_i,t2_r,t2_i,*pv,*Xpv;
  
  ierr = GetArray( &pv );           CHKERRQ(ierr);
  ierr = X->GetArray( &Xpv );       CHKERRQ(ierr);
  const double *Xvs = Xpv; double *vs = pv;
  double alpha_r = a.real(), alpha_i = a.imag(); 
  double beta_r = b.real(), beta_i = b.imag(); 
  for( ii=0 ; ii<sz ; ii++) {
    t1_r = *Xvs++; t1_i = *Xvs++;    t2_r = vs[0]; t2_i = vs[1]; 
    *vs++ = (alpha_r*t1_r - alpha_i*t1_i) + (beta_r*t2_r - beta_i*t2_i); 
    *vs++ = (alpha_r*t1_i + alpha_i*t1_r) + (beta_r*t2_i + beta_i*t2_r);
  }
  ierr = RestoreArray( &pv );       CHKERRQ(ierr);
  ierr = X->RestoreArray( &Xpv );   CHKERRQ(ierr);

  return 0;
}

#endif 
